"""
MTGJSON models
"""

import datetime
import sys

if sys.version_info < (3, 11):
    from typing import Dict, List, Optional, Union

    from typing_extensions import NotRequired, TypedDict
else:
    from typing import List, Optional, Union, TypedDict, Dict, NotRequired

from pydantic import Field
from pydantic.types import UUID

from mightstone.core import MightstoneModel


class Types(MightstoneModel):
    """
    The Types Data Model describes all types available on a Card.
    """

    sub_types: List[str] = Field(since="v4.0.0", alias="subTypes")
    """A list of all available subtypes of a type.
    Examples: "Abian", "Adventure", "Advisor", "Aetherborn", "Ajani" """
    super_types: List[str] = Field(since="v4.0.0", alias="superTypes")
    """A list of all available supertypes of a type.
    Examples: "Basic", "Host", "Legendary", "Ongoing", "Snow" """


class CardTypes(MightstoneModel):
    """
    The Card Types Data Model describes card types that a card may have.
    """

    artifact: Types = Field(since="v4.0.0", alias="artifact")
    """All possible subtypes and supertypes for Artifact cards."""

    conspiracy: Types = Field(since="v4.0.0", alias="conspiracy")
    """All possible subtypes and supertypes for Conspiracy cards."""

    creature: Types = Field(since="v4.0.0", alias="creature")
    """All possible subtypes and supertypes for Creature cards."""

    enchantment: Types = Field(since="v4.0.0", alias="enchantment")
    """All possible subtypes and supertypes for Enchantment cards."""

    instant: Types = Field(since="v4.0.0", alias="instant")
    """All possible subtypes and supertypes for Instant cards."""

    land: Types = Field(since="v4.0.0", alias="land")
    """All possible subtypes and supertypes for Land cards."""

    phenomenon: Types = Field(since="v4.0.0", alias="phenomenon")
    """All possible subtypes and supertypes for Phenomenon cards."""

    plane: Types = Field(since="v4.0.0", alias="plane")
    """All possible subtypes and supertypes for Plane cards."""

    planeswalker: Types = Field(since="v4.0.0", alias="planeswalker")
    """All possible subtypes and supertypes for Planeswalker."""

    scheme: Types = Field(since="v4.0.0", alias="scheme")
    """All possible subtypes and supertypes for Scheme cards."""

    sorcery: Types = Field(since="v4.0.0", alias="sorcery")
    """All possible subtypes and supertypes for Sorcery cards."""

    tribal: Types = Field(since="v4.0.0", alias="tribal")
    """All possible subtypes and supertypes for Tribal cards."""

    vanguard: Types = Field(since="v4.0.0", alias="vanguard")
    """All possible subtypes and supertypes for Vanguard cards."""


class DeckList(MightstoneModel):
    """
    The Deck List Data Model describes a metadata-like model for a Deck.
    """

    code: str = Field(since="v4.3.0", alias="code")
    """The set code for the deck."""

    file_name: str = Field(since="v4.3.0", alias="fileName")
    """The file name for the deck. Combines the name and code fields to avoid 
    namespace collisions and are given a delimiter of _. Examples: 
    "SpiritSquadron_VOC" """

    name: str = Field(since="v4.3.0", alias="name")
    """The name of the deck."""

    release_date: Optional[datetime.date] = Field(since="v4.3.0", alias="releaseDate")
    """The release date in ISO 8601 format for the set. Returns 
    null if the set was not formally released as a product. """

    type: str = Field(since="v4.3.0", alias="type")
    """The type of deck. Examples: "Advanced Deck", "Advanced Pack", "Archenemy 
    Deck", "Basic Deck", "Brawl Deck" """


class ForeignData(MightstoneModel):
    """
    The Foreign Data Data Model describes a list of properties for various Card Data
    Models in alternate languages.
    """

    face_name: Optional[str] = Field(since="v5.0.1", alias="faceName")
    """The foreign name on the face of the card."""

    flavor_text: Optional[str] = Field(since="v4.0.0", alias="flavorText")
    """The foreign flavor text of the card."""

    language: str = Field(since="v4.0.0", alias="language")
    """The foreign language of card. Examples: "Ancient Greek", "Arabic", "Chinese 
    Simplified", "Chinese Traditional", "French" """

    multiverse_id: Optional[int] = Field(since="v4.0.0", alias="multiverseId")
    """The foreign multiverse identifier of the card."""

    name: str = Field(since="v4.0.0", alias="name")
    """The foreign name of the card."""

    text: Optional[str] = Field(since="v4.0.0", alias="text")
    """The foreign text ruling of the card."""

    type: Optional[str] = Field(since="v4.0.0", alias="type")
    """The foreign type of the card. Includes any supertypes and subtypes."""


class Identifiers(MightstoneModel):
    card_kingdom_etched_id: Optional[str] = Field(
        since="v5.2.0", alias="cardKingdomEtchedId"
    )
    """The Card Kingdom etched card identifier."""

    card_kingdom_foil_id: Optional[str] = Field(
        since="v5.0.0", alias="cardKingdomFoilId"
    )
    """The Card Kingdom foil card identifier."""

    card_kingdom_id: Optional[str] = Field(since="v5.0.0", alias="cardKingdomId")
    """The Card Kingdom card identifier."""

    cardsphere_id: Optional[str] = Field(since="v5.2.1", alias="cardsphereId")
    """The Cardsphere card identifier."""

    mcm_id: Optional[str] = Field(since="v4.4.0", alias="mcmId")
    """The Card Market card identifier."""

    mcm_meta_id: Optional[str] = Field(since="v4.4.0", alias="mcmMetaId")
    """The Card Market card meta identifier."""

    mtg_arena_id: Optional[str] = Field(since="v4.5.0", alias="mtgArenaId")
    """The Magic: The Gathering Arena card identifier."""

    mtgo_foil_id: Optional[str] = Field(since="v4.5.0", alias="mtgoFoilId")
    """The Magic: The Gathering Online card foil identifier."""

    mtgo_id: Optional[str] = Field(since="v4.5.0", alias="mtgoId")
    """The Magic: The Gathering Online card identifier."""

    mtgjson_v4_id: Optional[str] = Field(since="v5.0.0", alias="mtgjsonV4Id")
    """The universal unique identifier generated by MTGJSON. Each entry is unique. 
    Entries are for MTGJSON v4 uuid generation. """

    multiverse_id: Optional[str] = Field(since="v4.0.0", alias="multiverseId")
    """The Wizards of the Coast card identifier used in conjunction with Gatherer."""

    scryfall_id: Optional[str] = Field(since="v4.0.0", alias="scryfallId")
    """The universal unique identifier generated by Scryfall. Note that cards with 
    multiple faces are not unique. """

    scryfall_oracle_id: Optional[str] = Field(since="v4.3.1", alias="scryfallOracleId")
    """The unique identifier generated by Scryfall for this card's oracle identity. 
    This value is consistent across reprinted card editions, and unique among 
    different cards with the same name (tokens, Unstable variants, etc). """

    scryfall_illustration_id: Optional[str] = Field(
        since="v4.3.1", alias="scryfallIllustrationId"
    )
    """The unique identifier generated by Scryfall for the card artwork that remains 
    consistent across reprints. Newly spoiled cards may not have this field yet. """

    tcgplayer_product_id: Optional[str] = Field(
        since="v4.2.1", alias="tcgplayerProductId"
    )
    """The TCGplayer card identifier."""

    tcgplayer_etched_product_id: Optional[str] = Field(
        since="v5.2.0", alias="tcgplayerEtchedProductId"
    )
    """The TCGplayer etched card identifier."""


class Keywords(MightstoneModel):
    ability_words: List[str] = Field(since="v4.3.0", alias="abilityWords")
    """A list of ability words found in rules text on cards.
    Examples: "Adamant", "Addendum", "Alliance", "Battalion", "Bloodrush" """

    keyword_abilities: List[str] = Field(since="v4.3.0", alias="keywordAbilities")
    """A list of keyword abilities found in rules text on cards
    Examples: "Absorb", "Affinity", "Afflict", "Afterlife", "Aftermath" """

    keyword_actions: List[str] = Field(since="v4.3.0", alias="keywordActions")
    """A list of keyword actions found in rules text on cards.
    Examples: "Abandon", "Activate", "Adapt", "Amass", "Assemble" """


class LeadershipSkills(MightstoneModel):
    brawl: bool = Field(since="v4.5.1", alias="brawl")
    """If the card can be your commander in the Brawl format."""

    commander: bool = Field(since="v4.5.1", alias="commander")
    """If the card can be your commander in the Commander/EDH format."""

    oathbreaker: bool = Field(since="v4.5.1", alias="oathbreaker")
    """If the card can be your commander in the Oathbreaker format."""


class Legalities(MightstoneModel):
    brawl: Optional[str] = Field(since="v4.0.0", alias="brawl")
    """If the card is legal in the Brawl play format."""

    commander: Optional[str] = Field(since="v4.0.0", alias="commander")
    """If the card is legal in the Commander play format."""

    duel: Optional[str] = Field(since="v4.0.0", alias="duel")
    """If the card is legal in the Duel Commander play format."""

    future: Optional[str] = Field(since="v4.0.0", alias="future")
    """If the card is legal in the future for the Standard play format."""

    frontier: Optional[str] = Field(since="v4.0.0", alias="frontier")
    """If the card is legal in the Frontier play format."""

    gladiator: Optional[str] = Field(since="v5.2.0", alias="gladiator")
    """If the card is legal in the Gladiator play format."""

    historic: Optional[str] = Field(since="v5.1.0", alias="historic")
    """If the card is legal in the Historic play format."""

    historicbrawl: Optional[str] = Field(since="v5.2.0", alias="historicbrawl")
    """If the card is legal in the Historic Brawl play format."""

    legacy: Optional[str] = Field(since="v4.0.0", alias="legacy")
    """If the card is legal in the Legacy play format."""

    modern: Optional[str] = Field(since="v4.0.0", alias="modern")
    """If the card is legal in the Modern play format."""

    oldschool: Optional[str] = Field(since="v5.2.0", alias="oldschool")
    """If the card is legal in the Old School play format."""

    pauper: Optional[str] = Field(since="v5.2.0", alias="pauper")
    """If the card is legal in the Pauper play format."""

    paupercommander: Optional[str] = Field(since="v5.2.0", alias="paupercommander")
    """If the card is legal in the Pauper Commander play format."""

    penny: Optional[str] = Field(since="v4.0.0", alias="penny")
    """If the card is legal in the Penny Dreadful play format."""

    pioneer: Optional[str] = Field(since="v4.6.0", alias="pioneer")
    """If the card is legal in the Pioneer play format."""

    premodern: Optional[str] = Field(since="v5.2.0", alias="premodern")
    """If the card is legal in the Pre-Modern play format."""

    standard: Optional[str] = Field(since="v4.0.0", alias="standard")
    """If the card is legal in the Standard play format."""

    vintage: Optional[str] = Field(since="v4.0.0", alias="vintage")
    """If the card is legal in the Vintage play format."""


class Meta(MightstoneModel):
    date: datetime.date = Field(since="v4.0.0", alias="date")
    """The current release date in ISO 8601 format for the MTGJSON build."""

    version: str = Field(since="v4.0.0", alias="version")
    """The current SemVer version for the MTGJSON build appended with the build date."""


class PurchaseUrls(MightstoneModel):
    card_kingdom: Optional[str] = Field(since="v5.0.0", alias="cardKingdom")
    """The URL to purchase a product on Card Kingdom."""

    card_kingdom_etched: Optional[str] = Field(
        since="v5.2.0", alias="cardKingdomEtched"
    )
    """The URL to purchase an etched product on Card Kingdom."""

    card_kingdom_foil: Optional[str] = Field(since="v5.0.0", alias="cardKingdomFoil")
    """The URL to purchase a foil product on Card Kingdom."""

    cardmarket: Optional[str] = Field(since="v4.4.0", alias="cardmarket")
    """"The URL to purchase a product on Cardmarket."""

    tcgplayer: Optional[str] = Field(since="v4.4.0", alias="tcgplayer")
    """The URL to purchase a product on TCGplayer."""

    tcgplayer_etched: Optional[str] = Field(since="v5.2.0", alias="tcgplayerEtched")
    """The URL to purchase an etched product on TCGplayer."""


class Rulings(MightstoneModel):
    date: Optional[datetime.date] = Field(since="v4.0.0", alias="date")
    """The release date in ISO 8601 format for the rule."""
    text: Optional[str] = Field(since="v4.0.0", alias="text")
    """The text ruling of the card."""


class SealedProduct(MightstoneModel):
    identifiers: Identifiers = Field(since="v5.2.0", alias="identifiers")
    """A list of identifiers associated to a product. See the Identifiers Data Model."""
    name: str = Field(since="v5.2.0", alias="name")
    """The name of the product."""
    purchase_urls: PurchaseUrls = Field(since="v5.2.0", alias="purchaseUrls")
    """Links that navigate to websites where the product can be purchased. See the 
    Purchase Urls Data Model. """
    release_date: Optional[datetime.date] = Field(since="v5.2.0", alias="releaseDate")
    """The release date in ISO 8601 format for the product."""
    uuid: UUID = Field(since="v5.2.0", alias="uuid")
    """The universal unique identifier (v5) generated by MTGJSON. Each entry is 
    unique. """


class TcgPlayerSKU(MightstoneModel):
    condition: str = Field(since="v5.1.0", alias="condition")
    """The condition of the card. Examples: "DAMAGED", "HEAVILY_PLAYED", 
    "LIGHTLY_PLAYED", "MODERATELY_PLAYED", "NEAR_MINT" """

    finishes: Optional[List[str]] = Field(since="v5.2.0", alias="finishes")
    """The finishes of the card.
    Examples: "FOIL_ETCHED" """

    language: str = Field(since="v5.1.0", alias="language")
    """The language of the card. Examples: "CHINESE_SIMPLIFIED", 
    "CHINESE_TRADITIONAL", "ENGLISH", "FRENCH", "GERMAN" """

    printing: str = Field(since="v5.1.0", alias="printing")
    """The printing style of the card.
    Examples: "FOIL", "NON_FOIL" """

    product_id: str = Field(since="v5.1.0", alias="productId")
    """The product identifier of the card."""

    sku_id: str = Field(since="v5.1.0", alias="skuId")
    """The SKU identifier of the card."""


class Translations(MightstoneModel):
    """
    The Translations Data Model describes a Set name translated per available language.
    """

    ancient_greek: Optional[str] = Field(since="v4.6.0", alias="Ancient Greek")
    """The set name translation in Ancient Greek."""

    arabic: Optional[str] = Field(since="v4.6.0", alias="Arabic")
    """The set name translation in Arabic."""

    chineese_simplified: Optional[str] = Field(
        since="v4.0.0", alias="Chinese Simplified"
    )
    """The set name translation in Chinese Simplified."""

    chineese_traditional: Optional[str] = Field(
        since="v4.6.0", alias="Chinese Traditional"
    )
    """The set name translation in Chinese Traditional."""

    french: Optional[str] = Field(since="v4.0.0", alias="French")
    """The set name translation in French."""

    german: Optional[str] = Field(since="v4.0.0", alias="German")
    """The set name translation in German."""

    hebrew: Optional[str] = Field(since="v4.6.0", alias="Hebrew")
    """The set name translation in Hebrew."""

    italian: Optional[str] = Field(since="v4.0.0", alias="Italian")
    """The set name translation in Italian."""

    japanese: Optional[str] = Field(since="v4.0.0", alias="Japanese")
    """The set name translation in Japanese."""

    korean: Optional[str] = Field(since="v4.0.0", alias="Korean")
    """The set name translation in Korean."""

    latin: Optional[str] = Field(since="v4.6.0", alias="Latin")
    """The set name translation in Latin."""

    phyrexian: Optional[str] = Field(since="v4.7.0", alias="Phyrexian")
    """The set name translation in Phyrexian."""

    portuguese_brazil: Optional[str] = Field(
        since="v4.0.0", alias="Portuguese (Brazil)"
    )
    """The set name translation in Portuguese (Brazil)."""

    russian: Optional[str] = Field(since="v4.0.0", alias="Russian")
    """The set name translation in Russian."""

    sanskrit: Optional[str] = Field(since="v4.6.0", alias="Sanskrit")
    """The set name translation in Sanskrit."""

    spanish: Optional[str] = Field(since="v4.0.0", alias="Spanish")
    """The set name translation in Spanish."""


class CardAtomic(MightstoneModel):
    """
    The Card (Atomic) Data Model describes the properties of a single atomic card,
    an oracle-like entity of a Magic: The Gathering card that only stores evergreen
    data that would never change from printing to printing.
    """

    ascii_name: Optional[str] = Field(since="v5.0.0", alias="asciiName")
    """The ASCII (Basic/128) code formatted card name with no 
    special unicode characters. """

    color_identity: List[str] = Field(since="v4.0.0", alias="colorIdentity")
    """A list of all the colors found in manaCost, colorIndicator, and text.
    Examples: "B", "G", "R", "U", "W" """

    color_indicator: Optional[List[str]] = Field(since="v4.0.2", alias="colorIndicator")
    """A list of all the colors in the color indicator (The symbol prefixed to a 
    card's types). Examples: "B", "G", "R", "U", "W" """

    colors: List[str] = Field(since="v4.0.0", alias="colors")
    """A list of all the colors in manaCost and colorIndicator. Some cards may not 
    have values, such as cards with "Devoid" in its text. Examples: "B", "G", "R", 
    "U", "W" """

    converted_mana_cost: float = Field(
        since="v4.0.0", alias="convertedManaCost", deprecated=True
    )
    """The converted mana cost of the card. Use the manaValue property."""

    edhrec_rank: Optional[int] = Field(since="v4.5.0", alias="edhrecRank")
    """The card rank on EDHRec."""

    face_converted_mana_cost: Optional[float] = Field(
        since="v4.1.1", alias="faceConvertedManaCost", deprecated=True
    )
    """The converted mana cost or mana value for the face for either half or part of 
    the card. Use the faceManaValue property. """

    face_mana_value: Optional[float] = Field(since="v5.2.0", alias="faceManaValue")
    """The mana value of the face for either half or part of the card. Formally known 
    as "converted mana cost". """

    face_name: Optional[str] = Field(since="v5.0.0", alias="faceName")
    """The name on the face of the card."""

    foreign_data: List[ForeignData] = Field(since="v4.0.0", alias="foreignData")
    """A list of data properties in other languages. See the Foreign Data Data Model."""

    hand: Optional[str] = Field(since="v4.2.1", alias="hand")
    """The starting maximum hand size total modifier. A + or - character precedes an 
    integer. """

    has_alternative_deck_limit: Optional[bool] = Field(
        since="v5.0.0", alias="hasAlternativeDeckLimit"
    )
    """If the card allows a value other than 4 copies in a deck."""

    identifiers: Identifiers = Field(since="v5.0.0", alias="identifiers")
    """A list of identifiers associated to a card. See the Identifiers Data Model."""

    is_funny: Optional[bool] = Field(since="v5.2.0", alias="isFunny")
    """If the card is part of a funny set."""
    is_reserved: Optional[bool] = Field(since="v4.0.1", alias="isReserved")
    """If the card is on the Magic: The Gathering Reserved List."""
    keywords: Optional[List[str]] = Field(since="v5.1.0", alias="keywords")
    """A list of keywords found on the card."""

    layout: str = Field(since="v4.0.0", alias="layout")
    """The type of card layout. For a token card, this will be "token".
    Examples: "adventure", "aftermath", "art_series", "augment", "class" """

    leadership_skills: Optional[LeadershipSkills] = Field(
        since="v4.5.1", alias="leadershipSkills"
    )
    """A list of formats the card is legal to be a commander in. See the Leadership 
    Skills Data Model. """

    legalities: Legalities = Field(since="v4.0.0", alias="legalities")
    """A list of play formats the card the card is legal in. See the Legalities Data 
    Model. """

    life: Optional[str] = Field(since="v4.2.1", alias="life")
    """The starting life total modifier. A plus or minus character precedes an 
    integer. Used only on cards with "Vanguard" in its types. """

    loyalty: Optional[str] = Field(since="v4.0.0", alias="loyalty")
    """The starting loyalty value of the card. Used only on cards with "Planeswalker" 
    in its types. """

    mana_cost: Optional[str] = Field(since="v4.0.0", alias="manaCost")
    """The mana cost of the card wrapped in brackets for each value.
    Example: "{1}{B}" """

    mana_value: float = Field(since="v5.2.0", alias="manaValue")
    """The mana value of the card. Formally known as "converted mana cost"."""

    name: str = Field(since="v4.0.0", alias="name")
    """The name of the card. Cards with multiple faces, like "Split" and "Meld" cards 
    are given a delimiter. Example: "Wear // Tear" """

    power: Optional[str] = Field(since="v4.0.0", alias="power")
    """The power of the card."""

    printings: Optional[List[str]] = Field(since="v4.0.0", alias="printings")
    """A list of set printing codes the card was printed in, formatted in uppercase."""

    purchase_urls: PurchaseUrls = Field(since="v4.4.0", alias="purchaseUrls")
    """Links that navigate to websites where the card can be purchased. See the 
    Purchase Urls Data Model. """

    rulings: List[Rulings] = Field(since="v4.0.0", alias="rulings")
    """The official rulings of the card. See the Rulings Data Model."""

    side: Optional[str] = Field(since="v4.1.0", alias="side")
    """The identifier of the card side. Used on cards with multiple faces on the same 
    card. Examples: "a", "b", "c", "d", "e" """

    subtypes: List[str] = Field(since="v4.0.0", alias="subtypes")
    """A list of card subtypes found after em-dash.
    Examples: "Abian", "Adventure", "Advisor", "Aetherborn", "Ajani" """

    supertypes: List[str] = Field(since="v4.0.0", alias="supertypes")
    """A list of card supertypes found before em-dash.
    Examples: "Basic", "Host", "Legendary", "Ongoing", "Snow" """

    text: Optional[str] = Field(since="v4.0.0", alias="text")
    """The rules text of the card."""

    toughness: Optional[str] = Field(since="v4.0.0", alias="toughness")
    """The toughness of the card."""

    type: str = Field(since="v4.0.0", alias="type")
    """The type of the card as visible, including any supertypes and subtypes."""

    types: List[str] = Field(since="v4.0.0", alias="types")
    """A list of all card types of the card, including Un‑sets and gameplay variants.
    Examples: "Artifact", "Card", "Conspiracy", "Creature", "Dragon" """


class CardDeck(MightstoneModel):
    artist: Optional[str] = Field(since="v4.0.0", alias="artist")
    """The name of the artist that illustrated the card art."""

    ascii_name: Optional[str] = Field(since="v5.0.0", alias="asciiName")
    """The ASCII (Basic/128) code formatted card name with no special unicode 
    characters. """

    availability: List[str] = Field(since="v5.0.0", alias="availability")
    """A list of the card's available printing types.
    Examples: "arena", "dreamcast", "mtgo", "paper", "shandalar" """

    booster_types: Optional[List[str]] = Field(since="v5.2.1", alias="boosterTypes")
    """A list of types this card is in a booster pack.
    Examples: "deck", "draft" """

    border_color: str = Field(since="v4.0.0", alias="borderColor")
    """The color of the card border.
    Examples: "black", "borderless", "gold", "silver", "white" """

    card_parts: Optional[List[str]] = Field(since="v5.2.0", alias="cardParts")
    """A list of card names associated to this card, such as "Meld" card face names."""

    color_identity: Optional[List[str]] = Field(since="v4.0.0", alias="colorIdentity")
    """A list of all the colors found in manaCost, colorIndicator, and text.
    Examples: "B", "G", "R", "U", "W" """

    color_indicator: Optional[List[str]] = Field(since="v4.0.2", alias="colorIndicator")
    """A list of all the colors in the color indicator (The symbol prefixed to a 
    card's types). Examples: "B", "G", "R", "U", "W" """

    colors: List[str] = Field(since="v4.0.0", alias="colors")
    """A list of all the colors in manaCost and colorIndicator. Some cards may not 
    have values, such as cards with "Devoid" in its text. Examples: "B", "G", "R", 
    "U", "W" """

    converted_mana_cost: float = Field(
        since="v4.0.0", alias="convertedManaCost", deprecated=True
    )
    """The converted mana cost of the card. Use the manaValue property."""

    count: int = Field(since="v4.4.1", alias="count")
    """The count of how many of this card exists in a relevant deck."""

    duel_deck: Optional[str] = Field(since="v4.2.0", alias="duelDeck")
    """The indicator for which duel deck the card is in.
    Examples: "a", "b" """

    edhrec_rank: Optional[int] = Field(since="v4.5.0", alias="edhrecRank")
    """The card rank on EDHRec."""

    face_converted_mana_cost: Optional[float] = Field(
        since="v4.1.1", alias="faceConvertedManaCost", deprecated=True
    )
    """The converted mana cost or mana value for the face for either half or part of 
    the card. Use the faceManaValue property. """

    face_flavor_name: Optional[str] = Field(since="v5.2.0", alias="faceFlavorName")
    """The flavor name on the face of the card."""

    face_mana_value: Optional[float] = Field(since="v5.2.0", alias="faceManaValue")
    """The mana value of the face for either half or part of the card. Formally known 
    as "converted mana cost". """

    face_name: Optional[str] = Field(since="v5.0.0", alias="faceName")
    """The name on the face of the card."""

    finishes: List[str] = Field(since="v5.2.0", alias="finishes")
    """The finishes of the card.
    Examples: "etched", "foil", "nonfoil", "signed" """

    flavor_name: Optional[str] = Field(since="v5.0.0", alias="flavorName")
    """The promotional card name printed above the true card name on special cards 
    that has no game function. See this card for an example. """

    flavor_text: Optional[str] = Field(since="v4.0.0", alias="flavorText")
    """The italicized text found below the rules text that has no game function."""

    foreign_data: List[ForeignData] = Field(since="v4.0.0", alias="foreignData")
    """A list of data properties in other languages. See the Foreign Data Data Model."""

    frame_effects: Optional[List[str]] = Field(since="v4.6.0", alias="frameEffects")
    """The visual frame effects.
    Examples: "colorshifted", "companion", "compasslanddfc", "convertdfc", "devoid"
    """
    frame_version: str = Field(since="v4.0.0", alias="frameVersion")
    """The version of the card frame style.
    Examples: "1993", "1997", "2003", "2015", "future" """

    hand: Optional[str] = Field(since="v4.2.1", alias="hand")
    """The starting maximum hand size total modifier. A + or - character precedes an 
    integer. """

    has_alternative_deck_limit: Optional[bool] = Field(
        since="v5.0.0", alias="hasAlternativeDeckLimit"
    )
    """If the card allows a value other than 4 copies in a deck."""

    has_content_warning: Optional[bool] = Field(
        since="v5.0.0", alias="hasContentWarning"
    )
    """If the card marked by Wizards of the Coast for having sensitive content. Cards 
    with this property may have missing or degraded properties and values. See this 
    official article for more information. """

    has_foil: bool = Field(since="v4.0.0", alias="hasFoil", deprecated=True)
    """If the card can be found in foil. Use the finishes property."""

    has_non_foil: bool = Field(since="v4.0.0", alias="hasNonFoil", deprecated=True)
    """If the card can be found in non-foil. Use the finishes property."""

    identifiers: Identifiers = Field(since="v5.0.0", alias="identifiers")
    """A list of identifiers associated to a card. See the Identifiers Data Model."""

    is_alternative: Optional[bool] = Field(since="v4.2.0", alias="isAlternative")
    """If the card is an alternate variation to an original printing."""

    is_foil: bool = Field(since="v5.0.0", alias="isFoil")
    """If the card is in foil."""

    is_full_art: Optional[bool] = Field(since="v4.4.2", alias="isFullArt")
    """If the card has full artwork."""

    is_funny: Optional[bool] = Field(since="v5.2.0", alias="isFunny")
    """If the card is part of a funny set."""

    is_online_only: Optional[bool] = Field(since="v4.0.1", alias="isOnlineOnly")
    """If the card is only available in online game variations."""

    is_oversized: Optional[bool] = Field(since="v4.0.0", alias="isOversized")
    """If the card is oversized."""

    is_promo: Optional[bool] = Field(since="v4.4.2", alias="isPromo")
    """If the card is a promotional printing."""

    is_rebalanced: Optional[bool] = Field(since="v5.2.0", alias="isRebalanced")
    """If the card is rebalanced for the Alchemy play format."""

    is_reprint: Optional[bool] = Field(since="v4.4.2", alias="isReprint")
    """If the card has been reprinted."""

    is_reserved: Optional[bool] = Field(since="v4.0.1", alias="isReserved")
    """If the card is on the Magic: The Gathering Reserved List."""

    is_starter: Optional[bool] = Field(since="v4.0.0", alias="isStarter")
    """If the card is found in a starter deck such as Planeswalker/Brawl decks."""

    is_story_spotlight: Optional[bool] = Field(since="v4.4.2", alias="isStorySpotlight")
    """If the card is a Story Spotlight card."""

    is_textless: Optional[bool] = Field(since="v4.4.2", alias="isTextless")
    """If the card does not have a text box."""

    is_timeshifted: Optional[bool] = Field(since="v4.4.1", alias="isTimeshifted")
    """If the card is "timeshifted", a feature of certain sets where a card will have 
    a different frameVersion. """

    keywords: Optional[List[str]] = Field(since="v5.0.0", alias="keywords")
    """A list of keywords found on the card."""

    language: str = Field(since="v5.2.1", alias="language")
    """The language the card is printed in. Examples: "Ancient Greek", "Arabic", 
    "Chinese Simplified", "Chinese Traditional", "English" """
    layout: str = Field(since="v4.0.0", alias="layout")
    """The type of card layout. For a token card, this will be "token".
    Examples: "adventure", "aftermath", "art_series", "augment", "class"
    """

    leadership_skills: Optional[LeadershipSkills] = Field(
        since="v4.5.1", alias="leadershipSkills"
    )
    """A list of formats the card is legal to be a commander in. See the Leadership 
    Skills Data Model. """

    legalities: Legalities = Field(since="v4.0.0", alias="legalities")
    """A list of play formats the card the card is legal in. See the Legalities Data 
    Model. """

    life: Optional[str] = Field(since="v4.2.1", alias="life")
    """The starting life total modifier. A plus or minus character precedes an 
    integer. Used only on cards with "Vanguard" in its types. """

    loyalty: Optional[str] = Field(since="v4.0.0", alias="loyalty")
    """The starting loyalty value of the card. Used only on cards with "Planeswalker" 
    in its types. """

    mana_cost: Optional[str] = Field(since="v4.0.0", alias="manaCost")
    """The mana cost of the card wrapped in brackets for each value.
    Example: "{1}{B}" """

    mana_value: float = Field(since="v5.2.0", alias="manaValue")
    """The mana value of the card. Formally known as "converted mana cost"."""

    name: str = Field(since="v4.0.0", alias="name")
    """The name of the card. Cards with multiple faces, like "Split" and "Meld" cards 
    are given a delimiter. Example: "Wear // Tear" """

    number: str = Field(since="v4.0.0", alias="number")
    """The number of the card. Can be prefixed or suffixed with a * or other 
    characters for promotional sets. """

    original_printings: Optional[List[str]] = Field(
        since="v5.2.0", alias="originalPrintings"
    )
    """A list of card UUID's to original printings of the card if this card is 
    somehow different from its original, such as rebalanced cards. """

    original_release_date: Optional[str] = Field(
        since="v5.1.0", alias="originalReleaseDate"
    )
    """The original release date in ISO 8601 format for a 
    promotional card printed outside of a cycle window, such as Secret Lair Drop 
    promotions. """

    original_text: Optional[str] = Field(since="v4.0.0", alias="originalText")
    """The text on the card as originally printed."""

    original_type: Optional[str] = Field(since="v4.0.0", alias="originalType")
    """The type of the card as originally printed. Includes any supertypes and 
    subtypes. """

    other_face_ids: Optional[List[str]] = Field(since="v4.6.1", alias="otherFaceIds")
    """A list of card UUID's to this card's counterparts, such as transformed or 
    melded faces. """

    power: Optional[str] = Field(since="v4.0.0", alias="power")
    """The power of the card. """

    printings: Optional[List[str]] = Field(since="v4.0.0", alias="printings")
    """A list of set printing codes the card was printed in, formatted in uppercase."""

    promo_types: Optional[List[str]] = Field(since="v5.0.0", alias="promoTypes")
    """A list of promotional types for a card.
    Examples: "alchemy", "arenaleague", "boosterfun", "boxtopper", "brawldeck" """

    purchaseUrls: PurchaseUrls = Field(since="v4.4.0", alias="purchaseUrls")
    """Links that navigate to websites where the card can be purchased. See the 
    Purchase Urls Data Model. """

    rarity: str = Field(since="v4.0.0", alias="rarity")
    """The card printing rarity. Rarity bonus relates to cards that have an alternate 
    availability in booster packs, while special relates to "Timeshifted" cards. 
    Examples: "bonus", "common", "mythic", "rare", "special" """

    rebalanced_printings: Optional[List[str]] = Field(
        since="v5.2.0", alias="rebalancedPrintings"
    )
    """A list of card UUID's to printings that are rebalanced 
    versions of this card. """

    rulings: List[Rulings] = Field(since="v4.0.0", alias="rulings")
    """The official rulings of the card. See the Rulings Data Model."""

    security_stamp: Optional[str] = Field(since="v5.2.0", alias="securityStamp")
    """The security stamp printed on the card.
    Examples: "acorn", "arena", "circle", "heart", "oval" """

    set_code: str = Field(since="v5.0.1", alias="setCode")
    """The set printing code that the card is from."""

    side: Optional[str] = Field(since="v4.1.0", alias="side")
    """The identifier of the card side. Used on cards with multiple faces on the same 
    card. Examples: "a", "b", "c", "d", "e" """

    signature: Optional[str] = Field(since="v5.2.0", alias="signature")
    """The name of the signature on the card."""

    subtypes: List[str] = Field(since="v4.0.0", alias="subtypes")
    """A list of card subtypes found after em-dash.
    Examples:
    "Abian", "Adventure", "Advisor", "Aetherborn", "Ajani" """

    supertypes: List[str] = Field(since="v4.0.0", alias="supertypes")
    """A list of card supertypes found before em-dash.
    Examples: "Basic", "Host", "Legendary", "Ongoing", "Snow" """

    text: Optional[str] = Field(since="v4.0.0", alias="text")
    """The rules text of the card."""

    toughness: Optional[str] = Field(since="v4.0.0", alias="toughness")
    """The toughness of the card."""

    type: str = Field(since="v4.0.0", alias="type")
    """The type of the card as visible, including any supertypes and subtypes."""

    types: List[str] = Field(since="v4.0.0", alias="types")
    """A list of all card types of the card, including Un‑sets and gameplay variants.
    Examples: "Artifact", "Card", "Conspiracy", "Creature", "Dragon" """

    uuid: UUID = Field(since="v4.0.0", alias="uuid")
    """The universal unique identifier (v5) generated by MTGJSON. Each entry is 
    unique. """

    variations: Optional[List[UUID]] = Field(since="v4.1.2", alias="variations")
    """A list of card UUID's of this card with alternate printings in the same set. 
    Excludes Un‑sets. """

    watermark: Optional[str] = Field(since="v4.0.0", alias="watermark")
    """The name of the watermark on the card.
    Examples: "abzan", "agentsofsneak", "arena", "atarka", "azorius" """


class CardSet(MightstoneModel):
    """
    The Card (Set) Data Model describes the properties of a single card in a ``Set``
    Data Model.
    """

    artist: Optional[str] = Field(since="v4.0.0", alias="artist")
    """The name of the artist that illustrated the card art."""

    ascii_name: Optional[str] = Field(since="v5.0.0", alias="asciiName")
    """The ASCII (Basic/128) code formatted card name with no 
    special unicode characters. """

    availability: List[str] = Field(since="v5.0.0", alias="availability")
    """A list of the card's available printing types.
    Examples: "arena", "dreamcast", "mtgo", "paper", "shandalar"
    """

    booster_types: Optional[List[str]] = Field(since="v5.2.1", alias="boosterTypes")
    """A list of types this card is in a booster pack.
    Examples: "deck", "draft" """

    border_color: str = Field(since="v4.0.0", alias="borderColor")
    """The color of the card border.
    Examples: "black", "borderless", "gold", "silver", "white" """

    card_parts: Optional[List[str]] = Field(since="v5.2.0", alias="cardParts")
    """A list of card names associated to this card, such as "Meld" card face names."""

    color_identity: List[str] = Field(since="v4.0.0", alias="colorIdentity")
    """A list of all the colors found in manaCost, colorIndicator, and text.
    Examples: "B", "G", "R", "U", "W" """

    color_indicator: Optional[List[str]] = Field(since="v4.0.2", alias="colorIndicator")
    """A list of all the colors in the color indicator (The symbol prefixed to a 
    card's types). Examples: "B", "G", "R", "U", "W" """

    colors: List[str] = Field(since="v4.0.0", alias="colors")
    """A list of all the colors in manaCost and colorIndicator. Some cards may not 
    have values, such as cards with "Devoid" in its text. Examples: "B", "G", "R", 
    "U", "W" """

    converted_mana_cost: float = Field(
        since="v4.0.0", alias="convertedManaCost", deprecated=True
    )
    """The converted mana cost of the card. Use the manaValue property."""

    edhrec_rank: Optional[int] = Field(since="v4.5.0", alias="edhrecRank")
    """The card rank on EDHRec."""

    face_converted_mana_cost: Optional[float] = Field(
        since="v4.1.1", alias="faceConvertedManaCost", deprecated=True
    )
    """The converted mana cost or mana value for the face for either half or part of 
    the card. Use the faceManaValue property. """

    face_flavor_name: Optional[str] = Field(since="v5.2.0", alias="faceFlavorName")
    """The flavor name on the face of the card."""

    face_mana_value: Optional[float] = Field(since="v5.2.0", alias="faceManaValue")
    """The mana value of the face for either half or part of the card. Formally known 
    as "converted mana cost". """

    face_name: Optional[str] = Field(since="v5.0.0", alias="faceName")
    """The name on the face of the card."""

    finishes: List[str] = Field(since="v5.2.0", alias="finishes")
    """The finishes of the card.
    Examples: "etched", "foil", "nonfoil", "signed" """

    flavor_name: Optional[str] = Field(since="v5.0.0", alias="flavorName")
    """The promotional card name printed above the true card name on special cards 
    that has no game function. See this card for an example. """

    flavor_text: Optional[str] = Field(since="v4.0.0", alias="flavorText")
    """The italicized text found below the rules text that has no game function."""

    foreign_data: List[ForeignData] = Field(since="v4.0.0", alias="foreignData")
    """A list of data properties in other languages. See the Foreign Data Data Model."""

    frame_effects: Optional[List[str]] = Field(since="v4.6.0", alias="frameEffects")
    """The visual frame effects.
    Examples: "colorshifted", "companion", "compasslanddfc", "convertdfc", "devoid" """

    frame_version: str = Field(since="v4.0.0", alias="frameVersion")
    """The version of the card frame style.
    Examples: "1993", "1997", "2003", "2015", "future" """

    hand: Optional[str] = Field(since="v4.2.1", alias="hand")
    """The starting maximum hand size total modifier. A + or - character precedes an 
    integer. """

    has_alternative_deck_limit: Optional[bool] = Field(
        since="v5.0.0", alias="hasAlternativeDeckLimit"
    )
    """If the card allows a value other than 4 copies in a deck."""

    has_content_warning: Optional[bool] = Field(
        since="v5.0.0", alias="hasContentWarning"
    )
    """If the card marked by Wiz""ards of the Coast for having 
    sensitive content. Cards with this property may have missing or degraded 
    properties and values. See this official article for more 
    information."" """

    has_foil: bool = Field(since="v4.0.0", alias="hasFoil", deprecated=True)
    """If the card can be found in foil. Use the finishes property."""

    has_non_foil: bool = Field(since="v4.0.0", alias="hasNonFoil", deprecated=True)
    """If the card can be found in non-foil. Use the finishes property."""

    identifiers: Identifiers = Field(since="v5.0.0", alias="identifiers")
    """A list of identifiers associated to a card. See the Identifiers Data Model."""

    is_alternative: Optional[bool] = Field(since="v4.2.0", alias="isAlternative")
    """If the card is an alternate variation to an original printing."""

    is_full_art: Optional[bool] = Field(since="v4.4.2", alias="isFullArt")
    """If the card has full artwork."""

    is_funny: Optional[bool] = Field(since="v5.2.0", alias="isFunny")
    """If the card is part of a funny set."""

    is_online_only: Optional[bool] = Field(since="v4.0.1", alias="isOnlineOnly")
    """If the card is only available in online game variations."""

    is_oversized: Optional[bool] = Field(since="v4.0.0", alias="isOversized")
    """If the card is oversized."""

    is_promo: Optional[bool] = Field(since="v4.4.2", alias="isPromo")
    """If the card is a promotional printing."""

    is_rebalanced: Optional[bool] = Field(since="v5.2.0", alias="isRebalanced")
    """If the card is rebalanced for the Alchemy 
    play format. """

    is_reprint: Optional[bool] = Field(since="v4.4.2", alias="isReprint")
    """If the card has been reprinted."""

    is_reserved: Optional[bool] = Field(since="v4.0.1", alias="isReserved")
    """If the card is on the Magic: The Gathering Reserved List."""

    is_starter: Optional[bool] = Field(since="v4.0.0", alias="isStarter")
    """If the card is found in a starter deck such as Planeswalker/Brawl decks."""

    is_story_spotlight: Optional[bool] = Field(since="v4.4.2", alias="isStorySpotlight")
    """If the card is a Story Spotlight card."""

    is_textless: Optional[bool] = Field(since="v4.4.2", alias="isTextless")
    """If the card does not have a text box."""

    is_timeshifted: Optional[bool] = Field(since="v4.4.1", alias="isTimeshifted")
    """If the card is "timeshifted", a feature of certain sets where a card will have 
    a different frameVersion. """

    keywords: Optional[List[str]] = Field(since="v5.0.0", alias="keywords")
    """A list of keywords found on the card"""

    language: str = Field(since="v5.2.1", alias="language")
    """The language the card is printed in. Examples: "Ancient Greek", "Arabic", 
    "Chinese Simplified", "Chinese Traditional", "English" """

    layout: str = Field(since="v4.0.0", alias="layout")
    """The type of card layout. For a token card, this will be "token".
    Examples: "adventure", "aftermath", "art_series", "augment", "class" """

    leadershipSkills: Optional[LeadershipSkills] = Field(
        since="v4.5.1", alias="leadershipSkills"
    )
    """A list of formats the card is legal to be a commander in. See the Leadership 
    Skills Data Model. """

    legalities: Legalities = Field(since="v4.0.0", alias="legalities")
    """A list of play formats the card the card is legal in. See the Legalities Data 
    Model. """

    life: Optional[str] = Field(since="v4.2.1", alias="life")
    """The starting life total modifier. A plus or minus character precedes an 
    integer. Used only on cards with "Vanguard" in its types. """

    loyalty: Optional[str] = Field(since="v4.0.0", alias="loyalty")
    """The starting loyalty value of the card. Used only on cards with "Planeswalker" 
    in its types. """

    mana_cost: Optional[str] = Field(since="v4.0.0", alias="manaCost")
    """The mana cost of the card wrapped in brackets for each value.
    Example: "{1}{B}" """

    mana_value: float = Field(since="v5.2.0", alias="manaValue")
    """The mana value of the card. Formally known as "converted mana cost"."""

    name: str = Field(since="v4.0.0", alias="name")
    """The name of the card. Cards with multiple faces, like "Split" and "Meld" cards 
    are given a delimiter of //. Example: "Wear // Tear" """

    number: str = Field(since="v4.0.0", alias="number")
    """The number of the card. Can be prefixed or suffixed with a * or other 
    characters for promotional sets. """

    original_printings: Optional[List[str]] = Field(
        since="v5.2.0", alias="originalPrintings"
    )
    """A list of card UUID's to original printings of the card if this card is 
    somehow different from its original, such as rebalanced cards. """

    original_release_date: Optional[str] = Field(
        since="v5.1.0", alias="originalReleaseDate"
    )
    """The original release date in ISO 8601 format for a promotional card printed 
    outside of a cycle window, such as Secret Lair Drop promotions. """

    original_text: Optional[str] = Field(since="v4.0.0", alias="originalText")
    """The text on the card as originally printed."""

    original_type: Optional[str] = Field(since="v4.0.0", alias="originalType")
    """The type of the card as originally printed. Includes any supertypes and 
    subtypes. """

    other_face_ids: Optional[List[str]] = Field(since="v4.6.1", alias="otherFaceIds")
    """A list of card UUID's to this card's counterparts, such as transformed or 
    melded faces. """

    power: Optional[str] = Field(since="v4.0.0", alias="power")
    """The power of the card."""

    printings: Optional[List[str]] = Field(since="v4.0.0", alias="printings")
    """A list of set printing codes the card was printed in, formatted in uppercase."""

    promo_types: Optional[List[str]] = Field(since="v5.0.0", alias="promoTypes")
    """A list of promotional types for a card.
    Examples: "alchemy", "arenaleague", "boosterfun", "boxtopper", "brawldeck"
    """

    purchase_urls: PurchaseUrls = Field(since="v4.4.0", alias="purchaseUrls")
    """Links that navigate to websites where the card can be purchased. See the 
    Purchase Urls Data Model. """

    rarity: str = Field(since="v4.0.0", alias="rarity")
    """The card printing rarity. Rarity bonus relates to cards that have an alternate 
    availability in booster packs, while special relates to "Timeshifted" cards. 
    Examples: "bonus", "common", "mythic", "rare", "special" """

    rebalanced_printings: Optional[List[UUID]] = Field(
        since="v5.2.0", alias="rebalancedPrintings"
    )
    """A list of card UUID's to printings that are rebalanced versions of this card."""

    rulings: List[Rulings] = Field(since="v4.0.0", alias="rulings")
    """The official rulings of the card. See the Rulings Data Model."""

    security_stamp: Optional[str] = Field(since="v5.2.0", alias="securityStamp")
    """The security stamp printed on the card.
    Examples: "acorn", "arena", "circle", "heart", "oval" """

    set_code: str = Field(since="v5.0.1", alias="setCode")
    """The set printing code that the card is from."""

    side: Optional[str] = Field(since="v4.1.0", alias="side")
    """The identifier of the card side. Used on cards with multiple faces on the same 
    card. Examples: "a", "b", "c", "d", "e" """

    signature: Optional[str] = Field(since="v5.2.0", alias="signature")
    """The name of the signature on the card."""

    subtypes: List[str] = Field(since="v4.0.0", alias="subtypes")
    """A list of card subtypes found after em-dash.
    Examples: "Abian", "Adventure", "Advisor", "Aetherborn", "Ajani" """

    supertypes: List[str] = Field(since="v4.0.0", alias="supertypes")
    """A list of card supertypes found before em-dash.
    Examples: "Basic", "Host", "Legendary", "Ongoing", "Snow" """

    text: Optional[str] = Field(since="v4.0.0", alias="text")
    """The rules text of the card."""

    toughness: Optional[str] = Field(since="v4.0.0", alias="toughness")
    """The toughness of the card."""

    type: str = Field(since="v4.0.0", alias="type")
    """Type of the card as visible, including any supertypes and subtypes."""

    types: List[str] = Field(since="v4.0.0", alias="types")
    """A list of all card types of the card, including Un‑sets and gameplay variants.
    Examples: "Artifact", "Card", "Conspiracy", "Creature", "Dragon" """

    uuid: UUID = Field(since="v4.0.0", alias="uuid")
    """The universal unique identifier (v5) generated by MTGJSON. Each entry is 
    unique. """

    variations: Optional[List[UUID]] = Field(since="v4.1.2", alias="variations")
    """A list of card UUID's of this card with alternate printings in the same set. 
    Excludes Un‑sets. """

    watermark: Optional[str] = Field(since="v4.0.0", alias="watermark")
    """The name of the watermark on the card.
    Examples: "abzan", "agentsofsneak", "arena", "atarka", "azorius" """


class CardToken(MightstoneModel):
    artist: Optional[str] = Field(since="v4.0.0", alias="artist")
    """The name of the artist that illustrated the card art."""

    ascii_name: Optional[str] = Field(since="v5.0.0", alias="asciiName")
    """The ASCII (Basic/128) code formatted card name with no special unicode 
    characters. """

    availability: List[str] = Field(since="v5.0.0", alias="availability")
    """A list of the card's available printing types.
    Examples: "arena", "dreamcast", "mtgo", "paper", "shandalar" """

    booster_types: Optional[List[str]] = Field(since="v5.2.1", alias="boosterTypes")
    """A list of types this card is in a booster pack.
    Examples: "deck", "draft" """

    border_color: str = Field(since="v4.0.0", alias="borderColor")
    """The color of the card border.
    Examples: "black", "borderless", "gold", "silver", "white" """

    card_parts: Optional[List[str]] = Field(since="v5.2.0", alias="cardParts")
    """A list of card names associated to this card, such as "Meld" card face names."""

    color_identity: List[str] = Field(since="v4.0.0", alias="colorIdentity")
    """A list of all the colors found in manaCost, colorIndicator, and text.
    Examples: "B", "G", "R", "U", "W" """

    color_indicator: Optional[List[str]] = Field(since="v4.0.2", alias="colorIndicator")
    """A list of all the colors in the color indicator (The symbol prefixed to a 
    card's types). Examples: "B", "G", "R", "U", "W" """

    colors: List[str] = Field(since="v4.0.0", alias="colors")
    """A list of all the colors in manaCost and colorIndicator. Some cards may not 
    have values, such as cards with "Devoid" in its text. Examples: "B", "G", "R", 
    "U", "W" """

    face_name: Optional[str] = Field(since="v5.0.0", alias="faceName")
    """The name on the face of the card."""

    face_flavor_name: Optional[str] = Field(since="v5.2.0", alias="faceFlavorName")
    """The flavor name on the face of the card."""

    finishes: List[str] = Field(since="v5.2.0", alias="finishes")
    """The finishes of the card.
    Examples: "etched", "foil", "nonfoil", "signed" """

    flavor_text: Optional[str] = Field(since="v4.0.0", alias="flavorText")
    """The italicized text found below the rules text that has no game function."""

    frame_effects: Optional[List[str]] = Field(since="v4.6.0", alias="frameEffects")
    """The visual frame effects.
    Examples: "colorshifted", "companion", "compasslanddfc", "convertdfc", "devoid" """

    frame_version: str = Field(since="v4.0.0", alias="frameVersion")
    """The version of the card frame style.
    Examples: "1993", "1997", "2003", "2015", "future" """

    has_foil: bool = Field(since="v4.0.0", alias="hasFoil", deprecated=True)
    """If the card can be found in foil. Use the finishes property."""

    has_non_foil: bool = Field(since="v4.0.0", alias="hasNonFoil", deprecated=True)
    """If the card can be found in non-foil. Use the finishes property."""

    identifiers: Identifiers = Field(since="v5.0.0", alias="identifiers")
    """A list of identifiers associated to a card. See the Identifiers Data Model."""

    is_full_art: Optional[bool] = Field(since="v4.4.2", alias="isFullArt")
    """If the card has full artwork."""

    is_funny: Optional[bool] = Field(since="v5.2.0", alias="isFunny")
    """If the card is part of a funny set."""

    is_online_only: Optional[bool] = Field(since="v4.0.1", alias="isOnlineOnly")
    """If the card is only available in online game variations."""

    is_promo: Optional[bool] = Field(since="v4.4.2", alias="isPromo")
    """If the card is a promotional printing."""

    is_reprint: Optional[bool] = Field(since="v4.4.2", alias="isReprint")
    """If the card has been reprinted."""

    keywords: Optional[List[str]] = Field(since="v5.0.0", alias="keywords")
    """A list of keywords found on the card."""

    language: str = Field(since="v5.2.1", alias="language")
    """The language the card is printed in. Examples: "Ancient Greek", "Arabic", 
    "Chinese Simplified", "Chinese Traditional", "English" """

    layout: str = Field(since="v4.0.0", alias="layout")
    """The type of card layout. For a token card, this will be "token".
    Examples: "adventure", "aftermath", "art_series", "augment", "class" """

    loyalty: Optional[str] = Field(since="v4.0.0", alias="loyalty")
    """The starting loyalty value of the card. Used only on cards with "Planeswalker" 
    in its types. """

    name: str = Field(since="v4.0.0", alias="name")
    """The name of the card. Cards with multiple faces, like "Split" and "Meld" cards 
    are given a delimiter. Example: "Wear // Tear" """

    number: str = Field(since="v4.0.0", alias="number")
    """The number of the card. Can be prefixed or suffixed with a * or other 
    characters for promotional sets. """

    other_face_ids: Optional[List[str]] = Field(since="v4.6.1", alias="otherFaceIds")
    """A list of card UUID's to this card's counterparts, such as transformed or 
    melded faces. """

    power: Optional[str] = Field(since="v4.0.0", alias="power")
    """The power of the card."""

    promo_types: Optional[List[str]] = Field(since="v5.0.0", alias="promoTypes")
    """A list of promotional types for a card.
    Examples: "alchemy", "arenaleague", "boosterfun", "boxtopper", "brawldeck" """

    reverse_related: List[str] = Field(since="v4.0.0", alias="reverseRelated")
    """The names of the cards that produce this card."""

    security_stamp: Optional[str] = Field(since="v5.2.0", alias="securityStamp")
    """The security stamp printed on the card.
    Examples: "acorn", "arena", "circle", "heart", "oval" """

    set_code: str = Field(since="v5.0.1", alias="setCode")
    """The set printing code that the card is from."""

    side: Optional[str] = Field(since="v4.1.0", alias="side")
    """The identifier of the card side. Used on cards with multiple faces on the same 
    card. Examples: "a", "b", "c", "d", "e" """

    signature: Optional[str] = Field(since="v5.2.0", alias="signature")
    """The name of the signature on the card."""

    subtypes: List[str] = Field(since="v4.0.0", alias="subtypes")
    """A list of card subtypes found after em-dash.
    Examples: "Abian", "Adventure", "Advisor", "Aetherborn", "Ajani" """

    supertypes: List[str] = Field(since="v4.0.0", alias="supertypes")
    """A list of card supertypes found before em-dash.
    Examples: "Basic", "Host", "Legendary", "Ongoing", "Snow" """

    text: Optional[str] = Field(since="v4.0.0", alias="text")
    """The rules text of the card."""

    toughness: Optional[str] = Field(since="v4.0.0", alias="toughness")
    """The toughness of the card."""

    type: str = Field(since="v4.0.0", alias="type")
    """The type of the card as visible, including any supertypes and subtypes."""

    types: List[str] = Field(since="v4.0.0", alias="types")
    """A list of all card types of the card, including Un‑sets and gameplay variants.
    Examples: "Artifact", "Card", "Conspiracy", "Creature", "Dragon" """

    uuid: UUID = Field(since="v4.0.0", alias="uuid")
    """The universal unique identifier (v5) generated by MTGJSON. Each entry is 
    unique. """

    watermark: Optional[str] = Field(since="v4.0.0", alias="watermark")
    """The name of the watermark on the card.
    Examples: "abzan", "agentsofsneak", "arena", "atarka", "azorius" """


class Deck(MightstoneModel):
    """
    The Deck Data Model describes a complete deck reference.
    """

    code: str = Field(since="v4.3.0", alias="code")
    """The set code for the deck."""

    file_name: Optional[str] = Field(since="v4.3.0", alias="fileName")
    """The file name for the deck. Combines the name and code fields to avoid 
    namespace collisions and are given a delimiter of _. Examples: 
    "SpiritSquadron_VOC" """

    name: str = Field(since="v4.3.0", alias="name")
    """The name of the deck."""

    release_date: Optional[datetime.date] = Field(since="v4.3.0", alias="releaseDate")
    """The release date in ISO 8601 format for the set. Returns 
    null if the set was not formally released as a product. """

    type: str = Field(since="v4.3.0", alias="type")
    """The type of deck. Examples: "Advanced Deck", "Advanced Pack", "Archenemy 
    Deck", "Basic Deck", "Brawl Deck" """

    commander: Optional[List[CardDeck]] = Field(since="v5.1.0", alias="commander")
    """The card that is the Commander in this deck. See the Card (Deck) Data Model."""

    main_board: List[CardDeck] = Field(since="v4.3.0", alias="mainBoard")
    """The cards in the main-board. See the Card (Deck) Data Model."""

    side_board: List[CardDeck] = Field(since="v4.3.0", alias="sideBoard")
    """The cards in the side-board. See the Card (Deck) Data Model."""


class Set(MightstoneModel):
    baseSetSize: int = Field(since="v4.1.0", alias="baseSetSize")
    """The number of cards in the set. This will default to totalSetSize if not 
    available. Wizards of the Coast sometimes prints extra cards 
    beyond the set size into promos or supplemental products. """

    block: Optional[str] = Field(since="v4.0.0", alias="block")
    """The block name the set was in."""

    booster: Optional[dict] = Field(since="v5.0.0", alias="booster")
    """A breakdown of possibilities and weights of cards in a booster pack. See the 
    Booster abstract model. """

    cards: List[CardSet] = Field(since="v4.0.0", alias="cards")
    """The list of cards in the set. See the Card (Set) Data Model."""

    cardsphere_set_id: Optional[int] = Field(since="v5.2.1", alias="cardsphereSetId")
    """The Cardsphere set identifier."""

    code: str = Field(since="v4.0.0", alias="code")
    """The set code for the set."""

    code_v3: Optional[str] = Field(since="v4.2.1", alias="codeV3")
    """The alternate set code Wizards of the Coast uses for a select few duel deck 
    sets. """

    is_foreign_only: Optional[bool] = Field(since="v4.4.1", alias="isForeignOnly")
    """If the set is available only outside the United States of America."""

    is_foil_only: bool = Field(since="v4.0.0", alias="isFoilOnly")
    """If the set is only available in foil."""

    is_non_foil_only: Optional[bool] = Field(since="v5.0.0", alias="isNonFoilOnly")
    """If the set is only available in non-foil."""

    is_online_only: bool = Field(since="v4.0.0", alias="isOnlineOnly")
    """If the set is only available in online game variations."""

    is_paper_only: Optional[bool] = Field(since="v4.6.2", alias="isPaperOnly")
    """If the set is available only in paper."""

    is_partial_preview: Optional[bool] = Field(since="v4.4.2", alias="isPartialPreview")
    """If the set is still in preview (spoiled). Preview sets do not have complete 
    data. """

    keyrune_code: str = Field(since="v4.3.2", alias="keyruneCode")
    """The matching Keyrune code for set image icons."""

    mcm_id: Optional[int] = Field(since="v4.4.0", alias="mcmId")
    """The Magic Card Market set identifier."""

    mcmIdExtras: Optional[int] = Field(since="v5.1.0", alias="mcmIdExtras")
    """The split Magic Card Market set identifier if a set is printed in two sets. 
    This identifier represents the second set's identifier. """

    mcmName: Optional[str] = Field(since="v4.4.0", alias="mcmName")
    """The Magic Card Market set name."""

    mtgoCode: Optional[str] = Field(since="v4.0.0", alias="mtgoCode")
    """The set code for the set as it appears on Magic: The Gathering Online."""

    name: str = Field(since="v4.0.0", alias="name")
    """The name of the set."""

    parent_code: Optional[str] = Field(since="v4.3.0", alias="parentCode")
    """The parent set code for set variations like promotions, guild kits, etc."""

    release_date: datetime.date = Field(since="v4.0.0", alias="releaseDate")
    """The release date in ISO 8601 format for the set."""

    sealed_product: Optional[List[SealedProduct]] = Field(
        since="v5.2.0", alias="sealedProduct"
    )
    """The sealed product information for the set. See the Sealed Product Data Model."""

    tcgplayerGroupId: Optional[int] = Field(since="v4.2.1", alias="tcgplayerGroupId")
    """The group identifier of the set on TCGplayer."""

    tokens: List[CardToken] = Field(since="v4.0.0", alias="tokens")
    """The tokens available to the set. See the Card (Token) Data Model."""

    totalSetSize: int = Field(since="v4.1.0", alias="totalSetSize")
    """The total number of cards in the set, including promotional and related 
    supplemental products but excluding Alchemy modifications - however those cards 
    are included in the set itself. """

    translations: Translations = Field(since="v4.3.2", alias="translations")
    """The translated set name by language. See the Translations Data Model."""

    type: str = Field(since="v4.0.0", alias="type")
    """The expansion type of the set.
    Examples: "alchemy", "archenemy", "arsenal", "box", "commander" """


class SetList(MightstoneModel):
    """
    The Set List Data Model describes a metadata-like properties and values for an
    individual Set.
    """

    baseSetSize: int = Field(since="v4.1.0", alias="baseSetSize")
    """The number of cards in the set. This will default to totalSetSize if not 
    available. Wizards of the Coast sometimes prints extra cards beyond the set size 
    into promos or supplemental products. """

    block: Optional[str] = Field(since="v4.0.0", alias="block")
    """The block name the set was in."""

    code: str = Field(since="v4.0.0", alias="code")
    """The set code for the set."""

    code_v3: Optional[str] = Field(since="v4.2.1", alias="codeV3")
    """The alternate set code Wizards of the Coast uses for a select few duel deck 
    sets. """

    is_foreign_only: Optional[bool] = Field(since="v4.4.1", alias="isForeignOnly")
    """If the set is available only outside the United States of America."""

    is_foil_only: bool = Field(since="v4.0.0", alias="isFoilOnly")
    """If the set is only available in foil."""

    is_non_foil_only: Optional[bool] = Field(since="v5.0.0", alias="isNonFoilOnly")
    """If the set is only available in non-foil."""

    is_online_only: bool = Field(since="v4.0.0", alias="isOnlineOnly")
    """If the set is only available in online game variations."""

    is_paper_only: Optional[bool] = Field(since="v4.6.2", alias="isPaperOnly")
    """If the set is only available in paper."""

    is_partial_preview: Optional[bool] = Field(since="v4.4.2", alias="isPartialPreview")
    """If the set is still in preview (spoiled). Preview sets do not have complete
    data. """

    keyrune_code: str = Field(since="v4.3.2", alias="keyruneCode")
    """The matching Keyrune code for set image icons."""

    mcm_id: Optional[int] = Field(since="v4.4.0", alias="mcmId")
    """The Magic Card Market set identifier."""

    mcm_id_extras: Optional[int] = Field(since="v5.1.0", alias="mcmIdExtras")
    """The split Magic Card Market set identifier if a set is printed in two sets.
    This identifier represents the second set's identifier. """

    mcm_name: Optional[str] = Field(since="v4.4.0", alias="mcmName")
    """The Magic Card Market set name."""

    mtgo_code: Optional[str] = Field(since="v4.0.0", alias="mtgoCode")
    """The set code for the set as it appears on Magic: The Gathering Online."""

    name: str = Field(since="v4.0.0", alias="name")
    """The name of the set."""

    parent_code: Optional[str] = Field(since="v4.3.0", alias="parentCode")
    """The parent set code for set variations like promotions, guild kits, etc."""

    release_date: datetime.date = Field(since="v4.0.0", alias="releaseDate")
    """The release date in ISO 8601 format for the set."""

    sealed_product: Optional[List[SealedProduct]] = Field(
        since="v5.1.0", alias="sealedProduct"
    )
    """The sealed product information for the set. See the Sealed Product Data Model."""

    tcgplayer_group_id: Optional[int] = Field(since="v4.2.1", alias="tcgplayerGroupId")
    """The group identifier of the set on TCGplayer."""

    total_set_size: int = Field(since="v4.1.0", alias="totalSetSize")
    """The total number of cards in the set, including promos and related
    supplemental products. """

    translations: Translations = Field(since="v4.3.2", alias="translations")
    """The translated set name by language. See the Translations Data Model."""

    type: str = Field(since="v4.0.0", alias="type")
    """The expansion type of the set.
    Examples: "alchemy", "archenemy", "arsenal", "box", "commander" """


class Card(MightstoneModel):
    """
    A card either a Card from a set, or a token
    """

    __root__: Union[CardSet, CardToken]


class RetailPrices(TypedDict):
    buylist: NotRequired[Dict[str, Dict[datetime.date, float]]]
    currency: str
    retail: Dict[str, Dict[datetime.date, float]]


class CardPrices(MightstoneModel):
    """
    A representation of the abstract model for card prices in MTGJSON
    """

    uuid: UUID
    mtgo: Optional[Dict[str, RetailPrices]]
    paper: Optional[Dict[str, RetailPrices]]


class CardAtomicGroup(MightstoneModel):
    """
    A representation of a group of Atomic Card such as returned by every _atomic
    methods of MtgJson client
    """

    name: str
    prints: List[CardAtomic]


class TcgPlayerSKUs(MightstoneModel):
    """
    A representation of a TcgPlayerSKU list associated to a card unique ID
    """

    uuid: UUID
    skus: List[TcgPlayerSKU]



import zlib


def get_gzip_encoder(compress_level: int = -1):
    return zlib.compressobj(
        compress_level,  # level: 0-9
        zlib.DEFLATED,  # method: must be DEFLATED
        16 + zlib.MAX_WBITS,  # window size in bits:
        #   -15..-8: negate, suppress header
        #   8..15: normal
        #   16..30: subtract 16, gzip header
        zlib.DEF_MEM_LEVEL,  # mem level: 1..8/9
        0  # strategy:
        #   0 = Z_DEFAULT_STRATEGY
        #   1 = Z_FILTERED
        #   2 = Z_HUFFMAN_ONLY
        #   3 = Z_RLE
        #   4 = Z_FIXED
    )


def get_gzip_decoder():
    return zlib.decompressobj(16 + zlib.MAX_WBITS)  # see above



from typing import TypedDict

import click

from mightstone.ass import asyncio_run, stream_as_list
from mightstone.cli.utils import catch_service_error, pretty_print
from mightstone.containers import Container
from mightstone.services.scryfall import (
    CardIdentifierPath,
    CatalogType,
    RulingIdentifierPath,
    Scryfall,
)


class ScryfallObj(TypedDict):
    container: Container
    client: Scryfall
    format: str


@click.group()
@click.pass_obj
def scryfall(obj: ScryfallObj, **kwargs):
    obj["client"] = obj["container"].scryfall(**kwargs)


@scryfall.command(name="sets")
@click.pass_obj
@click.option("--limit", type=int)
@catch_service_error
def scryfall_sets(obj: ScryfallObj, **kwargs):
    pretty_print(stream_as_list(obj["client"].sets(**kwargs)), obj.get("format"))


@scryfall.command(name="set")
@click.pass_obj
@click.argument("id_or_code", type=str)
def scryfall_set(obj: ScryfallObj, **kwargs):
    pretty_print(asyncio_run(obj["client"].set(**kwargs)), obj.get("format"))


@scryfall.command()
@click.pass_obj
@click.argument("id", type=str)
@click.argument("type", type=click.Choice([t.value for t in CardIdentifierPath]))
@catch_service_error
def card(obj: ScryfallObj, **kwargs):
    pretty_print(asyncio_run(obj["client"].card(**kwargs)), obj.get("format"))


@scryfall.command()
@click.pass_obj
@click.argument("q", type=str)
@click.option("--limit", type=int, default=100)
@catch_service_error
def search(obj: ScryfallObj, **kwargs):
    pretty_print(stream_as_list(obj["client"].search(**kwargs)), obj.get("format"))


@scryfall.command()
@click.pass_obj
@click.argument("q", type=str)
@catch_service_error
def random(obj: ScryfallObj, **kwargs):
    pretty_print(asyncio_run(obj["client"].random(**kwargs)), obj.get("format"))


@scryfall.command()
@click.pass_obj
@click.argument("q", type=str)
@click.option("--exact", type=bool, is_flag=True)
@click.option("--set", type=str)
@catch_service_error
def named(obj: ScryfallObj, **kwargs):
    pretty_print(asyncio_run(obj["client"].named(**kwargs)), obj.get("format"))


@scryfall.command()
@click.pass_obj
@click.argument("q", type=str)
@click.option("--include_extras", type=bool, is_flag=True)
@catch_service_error
def autocomplete(obj: ScryfallObj, **kwargs):
    pretty_print(asyncio_run(obj["client"].autocomplete(**kwargs)), obj.get("format"))


class ScryfallIdentifier(click.ParamType):
    name = "identifier"

    def convert(self, value, param, ctx):
        item = {}
        for constraint in value.split(","):
            (key, value) = constraint.split(":", 1)
            item[key] = value
        return item


@scryfall.command()
@click.pass_obj
@click.argument("identifiers", nargs=-1, type=ScryfallIdentifier())
@catch_service_error
def collection(obj: ScryfallObj, **kwargs):
    """
    scryfall collection id:683a5707-cddb-494d-9b41-51b4584ded69 "name:Ancient tomb"
    "set:dmu,collector_number:150"

    :param obj:
    :param kwargs:
    :return:
    """
    pretty_print(stream_as_list(obj["client"].collection(**kwargs)), obj.get("format"))


@scryfall.command()
@click.pass_obj
@click.argument("id", type=str)
@click.argument("type", type=click.Choice([t.value for t in RulingIdentifierPath]))
@click.option("-l", "--limit", type=int)
@catch_service_error
def rulings(obj: ScryfallObj, **kwargs):
    pretty_print(stream_as_list(obj["client"].rulings(**kwargs)), obj.get("format"))


@scryfall.command()
@click.pass_obj
@click.option("-l", "--limit", type=int, required=False)
@catch_service_error
def symbols(obj: ScryfallObj, **kwargs):
    pretty_print(stream_as_list(obj["client"].symbols(**kwargs)), obj.get("format"))


@scryfall.command()
@click.pass_obj
@click.argument("cost", type=str)
@catch_service_error
def parse_mana(obj: ScryfallObj, **kwargs):
    pretty_print(asyncio_run(obj["client"].parse_mana(**kwargs)), obj.get("format"))


@scryfall.command()
@click.pass_obj
@click.argument("type", type=click.Choice([t.value for t in CatalogType]))
@catch_service_error
def catalog(obj: ScryfallObj, **kwargs):
    pretty_print(asyncio_run(obj["client"].catalog(**kwargs)), obj.get("format"))


@scryfall.command()
@click.pass_obj
@click.option("-l", "--limit", type=int, default=100)
@catch_service_error
def migrations(obj: ScryfallObj, **kwargs):
    pretty_print(stream_as_list(obj["client"].migrations(**kwargs)), obj.get("format"))


@scryfall.command()
@click.pass_obj
@click.argument("id", type=str)
@catch_service_error
def migration(obj: ScryfallObj, **kwargs):
    pretty_print(asyncio_run(obj["client"].migration(**kwargs)), obj.get("format"))



from typing import TypedDict

import click

from mightstone.ass import asyncio_run
from mightstone.cli.utils import pretty_print
from mightstone.containers import Container
from mightstone.services.cardconjurer import CardConjurer


class CardConjurerObj(TypedDict):
    container: Container
    client: CardConjurer
    format: str


@click.group()
@click.pass_obj
@click.option("--cache", type=int, default=0)
def cardconjurer(obj: CardConjurerObj, **kwargs):
    obj["client"] = obj["container"].card_conjurer(**kwargs)


@cardconjurer.command()
@click.pass_obj
@click.argument("url_or_path")
def card(obj: CardConjurerObj, **kwargs):
    pretty_print(asyncio_run(obj["client"].card(**kwargs)), obj.get("format"))


@cardconjurer.command()
@click.pass_obj
@click.argument("url_or_path")
def template(obj: CardConjurerObj, **kwargs):
    pretty_print(asyncio_run(obj["client"].template(**kwargs)), obj.get("format"))


@cardconjurer.command()
@click.pass_obj
@click.argument("url_or_path")
@click.argument("output", type=click.File("wb"))
@click.option("--asset-root-url", type=str)
def render(obj: CardConjurerObj, url_or_path, output, asset_root_url):
    async def run():
        card = await obj["client"].card(url_or_path)
        if asset_root_url:
            card.asset_root_url = asset_root_url
        await obj["client"].render(card, output)

    asyncio_run(run())



"""
Scryfall.com support classes
"""

from typing import Any, List, Union

import ijson
from httpx import HTTPStatusError
from pydantic.error_wrappers import ValidationError
from typing_extensions import AsyncGenerator, TypeVar

from mightstone.services import MightstoneHttpClient, ServiceError
from mightstone.services.scryfall.models import (
    BulkTagType,
    Card,
    CardIdentifierPath,
    Catalog,
    CatalogType,
    DirectionStrategy,
    Error,
    IdentifierCollectorNumberSet,
    IdentifierId,
    IdentifierIllustrationId,
    IdentifierMtgId,
    IdentifierMultiverseId,
    IdentifierName,
    IdentifierNameSet,
    IdentifierOracleId,
    ManaCost,
    Migration,
    Ruling,
    RulingIdentifierPath,
    ScryfallList,
    Set,
    SortStrategy,
    Symbol,
    Tag,
    UniqueStrategy,
)

T = TypeVar("T")


class Scryfall(MightstoneHttpClient):
    """
    Scryfall API client
    """

    base_url = "https://api.scryfall.com"

    async def get_bulk_tags(self, tag_type: BulkTagType) -> AsyncGenerator[Tag, None]:
        """
        Access the private tag repository

        This is an alpha feature, and could be removed later.
        :param tag_type: The tag type either oracle or illustration
        :return: A scryfall `Tag` instance async generator
        """
        tag_type = BulkTagType(tag_type)
        async with self.client.get(f"/private/tags/{tag_type}") as f:
            f.raise_for_status()
            async for current_tag in ijson.items_async(f.content, "data.item"):
                yield Tag.parse_obj(current_tag)

    async def get_bulk_data(self, bulk_type: str) -> AsyncGenerator[Card, None]:
        """
        Access the bulk cards
        This script uses ijson and should stream data on the fly

        See https://scryfall.com/docs/api/bulk-data for more informations
        :param bulk_type: A string describing the bulk export name
        :return:
        """
        bulk_types = []
        bulk = None
        async with self.client.get("/bulk-data") as f:
            f.raise_for_status()
            async for current_bulk in ijson.items_async(f.content, "data.item"):
                bulk_types.append(current_bulk.get("type"))
                if current_bulk.get("type") == bulk_type:
                    bulk = current_bulk

        if not bulk:
            raise IndexError(f"{bulk_type} bulk type not found in {bulk_types}")

        async with self.client.get(bulk.get("download_uri")) as f:
            f.raise_for_status()
            async for current_card in ijson.items_async(f.content, "item"):
                yield Card.parse_obj(current_card)

    async def card(
        self, id: str, type: CardIdentifierPath = CardIdentifierPath.SCRYFALL
    ) -> Card:
        """
        Returns a single card with a given ID, or by its code set / collector number

        Depending on the `type` value, one of the following endpoint will be reached:
         * /cards/:id
         * /cards/tcgplayer/:id
         * /cards/multiverse/:id
         * /cards/mtgo/:id
         * /cards/arena/:id
         * /cards/cardmarket/:id
         * /cards/:code/:number

        :param id: The requested `Card` identifier string, for code-number, please
        use / as separator (dmu/123) :param type: The card identifier, please refer
        to `CardIdentifierPath` enum :return A scryfall `Card` instance
        :param type: Type of id researched
        """
        type = CardIdentifierPath(type)
        path = f"/cards/{id}"
        if type.value and type != CardIdentifierPath.CODE_NUMBER:
            path = f"/cards/{type.value}/{id}"
        return await self._get_item(path, Card)

    async def random(self, q: str = None) -> Card:
        """
        Returns a single random Card object.

        This method will use:
        - /cards/random

        :param q: The optional parameter q supports the same fulltext search system
        that the main site uses. Providing q will filter the pool of cards before
        returning a random entry. :return A scryfall `Card` instance
        """
        params = {}
        if q:
            params["q"] = q
        return await self._get_item("/cards/random", Card, params=params)

    async def search(
        self,
        q: str,
        unique: UniqueStrategy = UniqueStrategy.CARDS,
        order: SortStrategy = SortStrategy.NAME,
        dir: DirectionStrategy = DirectionStrategy.AUTO,
        include_extras=False,
        include_multilingual=False,
        include_variations=False,
        limit: int = 100,
    ) -> AsyncGenerator[Card, None]:
        """
        Returns a List object containing Cards found using a fulltext search string.
        This string supports the same fulltext search system that the main site uses.

        :param unique: The strategy for omitting similar cards.
        :param order: The method to sort returned cards.
        :param dir: The direction to sort cards.
        :param include_extras: If true, extra cards (tokens, planes, etc) will be
               included. Equivalent to adding include:extras to the fulltext search.
        :param include_multilingual: If true, cards in every language supported by
               Scryfall will be included.
        :param include_variations: If true, rare care variants will be included,
        like the Hairy Runesword.
        :param q: A fulltext search query.
        :param limit: The number of item to return, please note that Mightstone
        wraps Scryfall pagination and streams the results
        :return A scryfall `Card` instance async generator
        """
        params = {
            "unique": UniqueStrategy(unique).value,
            "order": SortStrategy(order).value,
            "dir": DirectionStrategy(dir).value,
            "include_extras": str(include_extras),
            "include_multilingual": str(include_multilingual),
            "include_variations": str(include_variations),
            "q": q,
        }
        async for item in self._list(
            "/cards/search", params=params, model=Card, limit=limit
        ):
            yield item

    async def named(self, q: str, set: str = None, exact=True):
        """
        Returns a Card based on a name search string. This method is designed for
        building chat bots, forum bots, and other services that need card details
        quickly.

        If exact mode is on, a card with that exact name is returned. Otherwise,
        an Exception is raised because no card matches. If exact mode is off and a
        card name matches that string, then that card is returned. If not, a fuzzy
        search is executed for your card name. The server allows misspellings and
        partial words to be provided. For example: jac bele will match Jace Beleren.
        When fuzzy searching, a card is returned if the server is confident that you
        unambiguously identified a unique name with your string. Otherwise,
        an exception will be raised describing the problem: either more than 1 one
        card matched your search, or zero cards matched.

        Card names are case-insensitive and punctuation is optional (you can drop
        apostrophes and periods etc). For example: fIReBALL is the same as Fireball
        and smugglers copter is the same as Smuggler's Copter.

        :param q: The searched card name :param exact: Run a strict text research
        instead of a fuzzy search :param set: You may also provide a set code in the
        set parameter, in which case the name search and the returned card print will
        be limited to the specified set. :return A scryfall `Card` instance
        """
        params = {}
        if exact:
            params["exact"] = q
        else:
            params["fuzzy"] = q
        if set:
            params["set"] = set
        return await self._get_item("/cards/named", Card, params=params)

    async def autocomplete(self, q: str, include_extras=False):
        """
        Returns a Catalog object containing up to 20 full English card names that
        could be autocompletions of the given string parameter.

        This method is designed for creating assistive UI elements that allow users
        to free-type card names.
        The names are sorted with the nearest match first, highly favoring results
        that begin with your given string.

        Spaces, punctuation, and capitalization are ignored.

        If q is less than 2 characters long, or if no names match, the Catalog will
        contain 0 items (instead of returning any errors).

        :param q: The string to autocomplete.
        :param include_extras: If true, extra cards (tokens, planes, vanguards, etc)
               will be included.
        :return: A scryfall `Card` instance async generator
        """
        params = {"q": q}
        if include_extras:
            params["include_extras"] = include_extras
        return await self._get_item("/cards/autocomplete", Catalog, params=params)

    async def collection(
        self,
        identifiers: List[
            Union[
                IdentifierId,
                IdentifierName,
                IdentifierNameSet,
                IdentifierMtgId,
                IdentifierOracleId,
                IdentifierMultiverseId,
                IdentifierCollectorNumberSet,
                IdentifierIllustrationId,
            ]
        ],
    ):
        """
        Accepts a JSON array of card identifiers, and returns a List object with the
        collection of requested cards. A maximum of 75 card references may be
        submitted per request.

        :param identifiers: Each submitted card identifier must be a JSON object with
        one or more of the keys id, mtgo_id, multiverse_id, oracle_id,
        illustration_id, name, set, and collector_number :return: A scryfall `Card`
        instance async generator
        """
        async for item in self._list(
            "/cards/collection", Card, verb="POST", json={"identifiers": identifiers}
        ):
            yield item

    async def rulings(
        self,
        id: str,
        type: RulingIdentifierPath = RulingIdentifierPath.SCRYFALL,
        limit: int = None,
    ):
        """
        Returns a single card with the given ID.

        Depending on the `type` value, one of the following endpoint will be reached:
         * /cards/:id/rulings
         * /cards/multiverse/:id/rulings
         * /cards/mtgo/:id/rulings
         * /cards/arena/:id/rulings

        :param id: The requested `Card` identifier string. In the case of card-number,
        use set/number (separated by a slash, for instance dmu/123)
        :param type: The card identifier, please refer to `RulingIdentifierPath` enum
        :return A scryfall `Ruling` instance async generator
        """
        type = RulingIdentifierPath(type)
        path = f"/cards/{id}/rulings"
        if type.value and type != RulingIdentifierPath.CODE_NUMBER:
            path = f"/cards/{type.value}/{id}/rulings"

        async for item in self._list(path, Ruling, limit=limit):
            yield item

    async def symbols(self, limit: int = None):
        """
        Returns a List of all Card Symbols.

        :param limit: The number of item to return, please note that Mightstone
        wraps Scryfall pagination and streams the results
        :return: A scryfall `Symbol` instance async generator
        """
        async for item in self._list("/symbology", Symbol, limit=limit):
            yield item

    async def parse_mana(self, cost: str):
        """
        Parses the given mana cost parameter and returns Scryfall’s interpretation.

        The server understands most community shorthand for mana costs (such as 2WW
        for {2}{W}{W}). Symbols can also be out of order, lowercase, or have multiple
        colorless costs (such as 2{g}2 for {4}{G}).

        If part of the string could not be understood, the server will raise an Error
        object describing the problem.

        :return: A `ManaCost` instance
        """
        return await self._get_item(
            "/symbology/parse-mana", ManaCost, params={"cost": cost}
        )

    async def catalog(self, type: CatalogType):
        """
        A Catalog object contains an array of Magic datapoints (words, card values,
        etc). Catalog objects are provided by the API as aids for building other
        Magic software and understanding possible values for a field on Card objects.

        :param type: See `CatalogType` for more informations
        :return: A `Catalog` instance
        """
        type = CatalogType(type)
        return await self._get_item(f"/catalog/{type.value}", Catalog)

    async def migrations(self, limit: int = None):
        """
        For the vast majority of Scryfall’s database, Magic card entries are additive.
        We add new and upcoming cards as we learn about them and obtain images.

        In rare instances, Scryfall may discover that a card in our database does not
        really exist, or it has been deleted from a digital game permanently. In
        these situations, we provide endpoints to help you reconcile downstream data
        you may have synced or imported from Scryfall.

        Each migration has a provided migration_strategy:

        merge
        You should update your records to replace the given old Scryfall ID with the
        new ID. The old ID is being discarded, and an existing record should be
        used to replace all instances of it.

        delete
        The given UUID is being discarded, and no replacement data is being provided.
        This likely means the old records are fully invalid. This migration exists to
        provide evidence that cards were removed from Scryfall’s database.

        :param limit: The number of item to return, please note that Mightstone wraps
        Scryfall pagination and streams the results
        :return: A `Migration` instance async generator
        """
        async for item in self._list("/migrations", Migration, limit=limit):
            yield item

    async def migration(self, id: str):
        """
        Returns a single Card Migration with the given :id

        :return: A `Migration` instance
        """
        return await self._get_item(f"/migrations/{id}", Migration)

    async def sets(self, limit: int = None):
        """
        Returns a List object of all Sets on Scryfall.

        :param limit: The number of item to return, please note that Mightstone
        wraps Scryfall pagination and streams the results
        :return: A `Set` instance async generator
        """
        async for item in self._list("/sets", Set, limit=limit):
            yield item

    async def set(self, id_or_code: str = None) -> Set:
        """
        Returns a Set with the given set code.

        :param id_or_code: The code can be either the code or the mtgo_code or the
        scryfall UUID for the set.
        :return: A `Set` instance
        """
        return await self._get_item(f"/sets/{id_or_code}", Set)

    async def _get_item(self, path, model: T = None, **kwargs) -> Union[T, Any]:
        try:
            async with self.client.get(path, **kwargs) as f:
                if not f.ok:
                    error = Error.parse_obj(await f.json())
                    raise ServiceError(
                        message=error.details,
                        method=f.method,
                        url=f.real_url,
                        status=f.status,
                        data=error,
                    )
                data = await f.json()
                if model:
                    data = model.parse_obj(data)
                return data
        except ValidationError as e:
            raise ServiceError(
                message=f"Failed to validate {model} data, {e.errors()}",
                url=path,
                method="GET",
                status=None,
                data=e,
            )
        except HTTPStatusError as e:
            raise ServiceError(
                message="Failed to fetch data from Scryfall",
                url=e.request.url,
                method=e.request.method,
                status=e.response.status_code,
                data=Error.parse_raw(e.response.content),
            )

    async def _list(
        self, path, model: T = None, verb="GET", limit=None, **kwargs
    ) -> AsyncGenerator[T, None]:
        i = 0
        try:
            while True:
                f = await self.client.request(verb, path, **kwargs)
                if f.is_error:
                    raise ServiceError(
                        message="Failed to fetch data from Scryfall",
                        url=f.real_url,
                        status=f.status,
                        data=Error.parse_obj(f.json()),
                    )

                my_list = ScryfallList.parse_obj(f.json())
                for item in my_list.data:
                    if limit and i >= limit:
                        return
                    i += 1

                    if model:
                        yield model.parse_obj(item)
                    else:
                        yield item

                if not my_list.has_more:
                    return

                path = my_list.next_page.path + "?" + my_list.next_page.query
                await self._sleep()
        except ValidationError as e:
            raise ServiceError(
                message=f"Failed to validate {model} data for item #{i}, {e.errors()}",
                url=path,
                status=None,
                data=e,
            )
        except HTTPStatusError as e:
            raise ServiceError(
                message="Failed to fetch data from Scryfall",
                url=e.request.url,
                status=e.response.status_code,
                data=Error.parse_obj(e),
            )



from typing import Iterable, Optional

from mightstone.ass.compressor import AsyncFileObj


class AsyncReader(object):
    def __init__(
        self,
        afd: AsyncFileObj,
        columns=Optional[Iterable[str]],
        column_types=Optional[Iterable[str]],
        has_header=False,
        sep=",",
        eol="\n",
    ):
        self._afd = afd
        self._sep = sep
        self._eol = eol
        self._columns = columns
        self._column_types = column_types
        self._has_header = has_header

    async def __aenter__(self):
        return self

    def __aiter__(self):
        return self

    async def __aexit__(self, *exc):
        return None

    async def __anext__(self):
        next_line = await self._afd.__anext__()
        if next_line:
            return next_line.decode("utf-8").rstrip(self._eol).split(self._sep)
        else:
            raise StopAsyncIteration



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['mightstone',
 'mightstone.ass',
 'mightstone.ass.compressor',
 'mightstone.ass.compressor.codecs',
 'mightstone.assets',
 'mightstone.cli',
 'mightstone.rule',
 'mightstone.services',
 'mightstone.services.cardconjurer',
 'mightstone.services.edhrec',
 'mightstone.services.mtgjson',
 'mightstone.services.scryfall']

package_data = \
{'': ['*']}

install_requires = \
['aiofiles>=22.1.0,<23.0.0',
 'aiosqlite>=0.18.0,<0.19.0',
 'aiostream>=0.4.5,<0.5.0',
 'appdirs>=1.4.4,<2.0.0',
 'asyncstdlib>=3.10.5,<4.0.0',
 'beanie>=1.17.0,<2.0.0',
 'cairosvg>=2.6.0,<3.0.0',
 'click>=8.0.1,<9.0.0',
 'dependency-injector>=4.41.0,<5.0.0',
 'httpx-cache>=0.7.0,<0.8.0',
 'httpx>=0.23.3,<0.24.0',
 'ijson>=3.2.0.post0,<4.0.0',
 'logging>=0.4.9.6,<0.5.0.0',
 'nest-asyncio>=1.5.6,<2.0.0',
 'ordered-set>=4.1.0,<5.0.0',
 'pillow>=9.4.0,<10.0.0',
 'pytest>=7.2.1,<8.0.0',
 'python-slugify>=8.0.0,<9.0.0',
 'requests>=2.28.2,<3.0.0']

entry_points = \
{'console_scripts': ['mightstone = mightstone.cli:cli']}

setup_kwargs = {
    'name': 'mightstone',
    'version': '0.4.0',
    'description': 'A library manage all things Magic The Gathering related in python',
    'long_description': '# mightstone\n\n\n[![PyPi](https://img.shields.io/pypi/v/mightstone.svg)](https://pypi.python.org/pypi/mightstone)\n[![Documentation](https://readthedocs.org/projects/mightstone/badge/?version=latest)](https://mightstone.readthedocs.io/en/latest/?badge=latest)\n[![security: bandit](https://img.shields.io/badge/security-bandit-yellow.svg)](https://github.com/PyCQA/bandit)\n[![OpenSSF Best Practices](https://bestpractices.coreinfrastructure.org/projects/7037/badge)](https://bestpractices.coreinfrastructure.org/projects/7037)\n\nA library manage all things Magic The Gathering related in python\n\n\n## Developing\n\nRun `make` for help\n\n    make install             # Run `poetry install`\n    make showdeps            # run poetry to show deps\n    make lint                # Runs bandit and black in check mode\n    make format              # Formats you code with Black\n    make test                # run pytest with coverage\n    make build               # run `poetry build` to build source distribution and wheel\n    make pyinstaller         # Create a binary executable using pyinstaller\n\n\n## System dependencies\n\nMightstone use [Ijson](https://github.com/ICRAR/ijson) that relies on [YAJL](https://lloyd.github.io/yajl/). IJson will\nuse its python backend on the run if YAJL is not installed, but you cold benefit from installing YAJL locally.\n\n\n',
    'author': 'Guillaume Boddaert',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/guibod/mightstone',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9.0,<4.0.0',
}


setup(**setup_kwargs)



class NoneDecompressor(object):
    def decompress(self, data: bytes) -> bytes:
        return data

    def flush(self) -> bytes:
        return b""


class NoneCompressor(object):
    def compress(self, data: bytes) -> bytes:
        return data

    def flush(self) -> bytes:
        return b""



import logging

logger = logging.getLogger("mightstone")


class App:
    def __init__(self, settings):
        self.settings = settings

    def start(self):
        logger.info("mightstone Starting")

    def stop(self):
        logger.info("mightstone Stopping")



from typing import Any, Iterable, List, Optional

from mightstone.ass.compressor import AsyncFileObj


class AsyncWriter(object):
    def __init__(
        self,
        afd: AsyncFileObj,
        columns=Optional[Iterable[str]],
        column_types=Optional[Iterable[str]],
        has_header=False,
        sep=b",",
        eol=b"\n",
    ):
        self._afd = afd
        self._sep = sep
        self._eol = eol
        self._columns = columns
        self._column_types = column_types
        self._has_header = has_header

    async def __aenter__(self):
        return self

    async def __aexit__(self, *exc):
        return await self._afd.close()

    async def writeheader(self):
        await self._afd.write(self._sep.join(self._columns) + self._eol)

    async def writerow(self, row: List[Any]):
        await self._afd.write(self._sep.join(row) + self._eol)

    async def writerows(self, rows: List[List[Any]]):
        for row in rows:
            await self.writerow(row)

    async def flush(self):
        await self._afd.flush()



import httpx_cache
from appdirs import user_cache_dir
from dependency_injector import containers, providers

from .services.cardconjurer import CardConjurer
from .services.edhrec import EdhRecApi, EdhRecStatic
from .services.mtgjson import MtgJson
from .services.scryfall import Scryfall


class Container(containers.DeclarativeContainer):
    # Gateways
    httpx_cache_transport = providers.Factory(
        httpx_cache.AsyncCacheControlTransport,
        cache=httpx_cache.FileCache(cache_dir=user_cache_dir("mightstone")),
    )

    # Services
    scryfall = providers.Factory(
        Scryfall,
        transport=httpx_cache_transport,
    )

    edhrec_static = providers.Factory(
        EdhRecStatic,
        transport=httpx_cache_transport,
    )

    edhrec_api = providers.Factory(
        EdhRecApi,
        transport=httpx_cache_transport,
    )

    card_conjurer = providers.Factory(
        CardConjurer,
        transport=httpx_cache_transport,
    )

    mtg_json = providers.Factory(
        MtgJson,
        transport=httpx_cache_transport,
    )
